/* Receiving Component, as the name implies only deals with receving data. Same concept as the transmitter, it will subscribe for 
receving adapters from adapter manager, and use the ones that are available.

Note for enhancements in the future;
i) Logic to dynamically switch adapters, either based on their connection status or other factors
ii) Enabling the use of mutli adapters usage to increase bandwith for data transmission. (More Advanced)

 */
import { BehaviorSubject, filter, Observable, Observer, Subject, Subscription } from 'rxjs';
import { checkMessage, WrappedMessage } from '../utils/message.ordering';
import ConsoleLogger from '../utils/log.utils';
import { MessageTransmissionBase } from '../base/msg.transmission.base';
import { AdapterInterface, AdapterManagerInterface, ConnectionState, FisMessage, GeneralEvent, MessageReceiverInterface, ReceiverAdapterInterface, TransmissionProfile, TransportMessage } from '../interface/interface';

export class MessageTransmissionReceiver extends MessageTransmissionBase implements MessageReceiverInterface {
    private connectionStateEvent: BehaviorSubject<ConnectionState> = new BehaviorSubject<ConnectionState>(`OFFLINE`)
    private console: ConsoleLogger = new ConsoleLogger(`MessageTransmissionReceiver`, ['transmission'])
    private onHoldMessage: Subject<WrappedMessage> = new Subject()
    private currentAdapter!: ReceiverAdapterInterface
    private incomingMessage: Subject<GeneralEvent<TransportMessage>> = new Subject()

    constructor(profile: TransmissionProfile, adapterManager: AdapterManagerInterface) {
        super()
        this.profile = profile
        this.console.log({ message: `Constructing Receiver Transmission for Receiving target: ${this.profile.target}` })
        this.initializeReceiverComponents(adapterManager)
    }

    public getReceivables(): Observable<FisMessage> {
        return new Observable((receivable: Observer<FisMessage>) => {
            this.console.log({ message: `Tranmission Subscription: Streaming incoming messages from ${this.profile.target}` })
            const subscription: Subscription = this.incomingMessage.pipe(
                filter((event: GeneralEvent<any>) => event.type == `Adapter Event`),
                filter((event: GeneralEvent<any>) => event.event == 'New Message'),
            ).subscribe((event: GeneralEvent<TransportMessage>) => {
                this.onHoldMessage.next(((event.data as TransportMessage).payload as WrappedMessage))
                checkMessage(((event.data as TransportMessage).payload as WrappedMessage), this.onHoldMessage).then(() => {
                    // only release the message before it exists
                    this.console.log({ message: `This one passes. Does have previousID. Case for message ordering` })
                    receivable.next(((event.data as TransportMessage).payload as WrappedMessage).payload as FisMessage);
                }).catch((error) => {
                    this.console.log({ message: `Observer Error`, details: error })
                })
            })

            // Clean up on unsubscription
            return () => {
                subscription.unsubscribe();
            };
        })

    }

    /* Assigned and update adapters record. Currently no logic to swtich adapters based on performance or whatever logic to be integrated in the future */
    private initializeReceiverComponents(adapterManager: AdapterManagerInterface): void {
        adapterManager.subscribeForAdapters(this.profile.target, `Receiver`).pipe(
        ).subscribe((adapter: AdapterInterface) => {
            this.adapters.push(adapter)
            this.console.log({ message: `Adding new ${adapter.getAdapterProfile(`transportType`)} receiving adapter. Current adapter length: ${this.adapters.length}` })
            if (!this.currentAdapter) {
                this.console.log({ message: `Setting this ${adapter.getAdapterProfile(`id`)} as current adapter.` })
                this.currentAdapter = adapter as ReceiverAdapterInterface
                this.currentAdapter.subscribeForIncoming().subscribe({
                    next: (message: GeneralEvent<TransportMessage>) => {
                        this.console.log({ message: `Received ${(((message.data as TransportMessage).payload as WrappedMessage).payload as FisMessage).header.messageID} from ${((message.data as TransportMessage).source)}`, details: message })
                        this.incomingMessage.next(message)
                    },
                    error: error => {
                        // Error handling. Idealling switching to other adapters
                    }
                })
                let connectionState: Observable<ConnectionState> = this.currentAdapter.getAdapterProfile(`connectionState`) as Observable<ConnectionState>
                connectionState.subscribe(this.connectionStateEvent)
            } else {
                this.currentAdapter.subscribeForIncoming().subscribe({
                    next: (message: GeneralEvent<TransportMessage>) => this.incomingMessage.next(message),
                    error: error => {
                        // Error handling. Idealling switching to other adapters
                    }
                })
            }
        })
    }

}
